------------
-- LOCALS --
------------
local _

local mabs = math.abs
local mmax = math.max
local mfloor = math.floor
local tinsert = table.insert
local tremove = table.remove
local tconcat = table.concat
local function meq(a, b) return mabs(a - b) <= 0.01 end
local function mlte(a, b) return meq(a, b) or a < b end
local function mgte(a, b) return meq(a, b) or a > b end
local sformat = string.format
local sfind = string.find
local ssub = string.sub
local slen = string.len
local xand = bit.band
local tostring = tostring
local pairs = pairs
local ipairs = ipairs
local next = next
local _print = print
local debug = true
local print = function(...) if not debug then return end return _print(...) end
local printf = function(...) if not debug then return end return _print(string.format(...)) end

local CreateFrame = CreateFrame
local UIParent = UIParent
local UnitGUID = UnitGUID
local UnitIsDead = UnitIsDead
local UnitIsUnit = UnitIsUnit
local GetTime = GetTime
local GetSpellTexture = GetSpellTexture
local InterfaceOptionsFrame_OpenToCategory = InterfaceOptionsFrame_OpenToCategory
local GetCVar = GetCVar
local SetCVar = SetCVar
local C_NamePlate = C_NamePlate
local LibStub = LibStub
local COMBATLOG_OBJECT_TYPE_GUARDIAN = COMBATLOG_OBJECT_TYPE_GUARDIAN
local COMBATLOG_OBJECT_TYPE_PET = COMBATLOG_OBJECT_TYPE_PET
local COMBATLOG_OBJECT_AFFILIATION_MINE = COMBATLOG_OBJECT_AFFILIATION_MINE
local CombatLogGetCurrentEventInfo = CombatLogGetCurrentEventInfo

local function createMap()
    local map = {len = 0, cnt = 0, index = {}, container = {}}
    function map:count()
        return self.cnt
    end
    function map:iter()
        local i = 1
        local ret
        return function()
            while true do
                if i > self.len then
                    return
                end
                ret = self.container[i]
                if ret then break end
                i = i + 1
            end
            i = i + 1
            return ret.key, ret.val
        end
    end
    function map:emplace(key, val)
        if key == nil then return end
        local idx = self.index[key]
        if idx then
            self.container[idx].val = val
            return
        end
        self.len = self.len + 1
        self.container[self.len] = {key = key, val = val}
        self.index[key] = self.len
        self.cnt = self.cnt + 1
    end
    function map:remove(key)
        local idx = self.index[key]
        if not idx then return end
        local ret = self.container[idx]
        self.container[idx] = nil
        self.index[key] = nil
        self.cnt = self.cnt - 1
        return ret.val
    end
    function map:at(key)
        local idx = self.index[key]
        if not idx then return end
        return self.container[idx].val
    end
    function map:trunk()
        local container = {}
        local len = 0
        if self.cnt > 0 then
            for i = 1, self.len do
                local ret = self.container[i]
                if ret then
                    len = len + 1
                    container[len] = ret
                    self.index[ret.key] = len
                end
            end
        end
        self.container = container
        self.len = len
    end
    return map
end

local animating = createMap()

local playerGUID
local unitToGuid = {}
local guidToUnit = {}
local guidToAnim = {}


---------------
-- LIBRARIES --
---------------
local AceAddon = LibStub("AceAddon-3.0")
local LibEasing = LibStub("LibEasing-1.0")
local AceLocale = LibStub('AceLocale-3.0')
local SharedMedia = LibStub("LibSharedMedia-3.0")

local AceGUIWidgetLSMlists = AceGUIWidgetLSMlists
local L = AceLocale:GetLocale("ClassicNFCT")
L.UI = setmetatable(L.UI or {}, { __index = function(t, k) return k end, })

ClassicNFCT = AceAddon:NewAddon("ClassicNFCT", "AceConsole-3.0", "AceEvent-3.0")
local ClassicNFCT = ClassicNFCT

ClassicNFCT.frame = CreateFrame("Frame", nil, UIParent)


--------
-- DB --
--------
local defaultFont = "Friz Quadrata TT"
if (SharedMedia:IsValid("font", "Bazooka")) then
    defaultFont = "Bazooka"
end

local truncateWords = {"", "K", "M", "B"}
local fmtStyles = L.fmtStyles
local iconValues = L.iconValues
local fontFlags = L.fontFlags

local fontSize = 30

local defaults = {
    global = {
        enabled = true,
        blzDisabled = true,

        font = defaultFont,
        fontFlag = "",
        textShadow = true,
        damageColor = false,

        fmtStyle = "commaSep",

        animations = {
            animationspeed = 1,
            animMaxLines = 5,
            animationCount = 80,
        },

        distance = 25,

        formatting = {
            icon = "none",
            scale = 1,
            alpha = 1,
        },

        petFormatting = { scale = 0.7, },
        autoAttackFormatting = { scale = 0.7, },

        useOffTarget = true,
        offTargetFormatting = {
            scale = 0.5,
            alpha = 0.33,
        },
    },
}


---------------------
-- LOCAL CONSTANTS --
---------------------
local ANIMATION_HORIZONTAL_PADDING = 20
local ANIMATION_VERTICAL_DISTANCE = 5
local ANIMATION_CRIT_SCALE = 1.4
local ANIMATION_CRIT_MAX_SCALE = 1.7

local DAMAGE_TYPE_COLORS = {
    [SCHOOL_MASK_PHYSICAL] = "FFFF00",
    [SCHOOL_MASK_HOLY] = "FFE680",
    [SCHOOL_MASK_FIRE] = "FF8000",
    [SCHOOL_MASK_NATURE] = "4DFF4D",
    [SCHOOL_MASK_FROST] = "80FFFF",
    [SCHOOL_MASK_SHADOW] = "8080FF",
    [SCHOOL_MASK_ARCANE] = "FF80FF",
	[SCHOOL_MASK_FIRE + SCHOOL_MASK_FROST + SCHOOL_MASK_ARCANE + SCHOOL_MASK_NATURE + SCHOOL_MASK_SHADOW] = "A330C9", -- Chromatic
	[SCHOOL_MASK_FIRE + SCHOOL_MASK_FROST + SCHOOL_MASK_ARCANE + SCHOOL_MASK_NATURE + SCHOOL_MASK_SHADOW + SCHOOL_MASK_HOLY] = "A330C9", -- Magic
	[SCHOOL_MASK_PHYSICAL + SCHOOL_MASK_FIRE + SCHOOL_MASK_FROST + SCHOOL_MASK_ARCANE + SCHOOL_MASK_NATURE + SCHOOL_MASK_SHADOW + SCHOOL_MASK_HOLY] = "A330C9", -- Chaos
	["melee"] = "FFFFFF",
    ["pet"] = "CC8400",
    ["spell"] = "FFFF00",
}

local MISS_EVENT_STRINGS = L.MISS_EVENT_STRINGS


---------------
-- ANIMATION --
---------------

local lineHeight = fontSize * ANIMATION_CRIT_SCALE
local function createAnimation(guid)
    local anim = guidToAnim[guid]
    if anim then return anim end

    anim = {lines = {{}}, lineMaxEdge = 0, version = 1, width = 0, height = 0}
    
    guidToAnim[guid] = anim

    function anim:updateLineMaxEdge()
        local len = #(self.lines)
        self.height = len * lineHeight
        -- WIDTH : HEIGHT = 2 : 1
        self.lineMaxEdge = (len + 1) * 3 / 3 * lineHeight
    end
    
    function anim:updateWidth()
        local half = 0
        for _, line in ipairs(self.lines) do
            local leftSpace = line[1]
            if leftSpace then
                half = mmax(mabs(leftSpace.to), mabs(line[#line].from))
            end
        end
        self.width = half * 2
    end

    function anim:insertFontStringToLine(lineIdx, fontString, fontWidth, fontHalfWidth)
        local line = self.lines[lineIdx]
        if not line then
            tinsert(self.lines, {})
            tinsert(self.lines, {})
            self:updateLineMaxEdge()
            return self:insertFontStringToLine(1, fontString, fontWidth, fontHalfWidth)
        end

        if lineIdx > ClassicNFCT.db.global.animations.animMaxLines then
            self.lines = {{}}
            self.version = self.version + 1
            self:updateWidth()
            self:updateLineMaxEdge()
            return self:insertFontStringToLine(1, fontString, fontWidth, fontHalfWidth)
        end
    
        if not line[1] then
            line[1] = {to = -fontHalfWidth}
            line[2] = {fontString = fontString, from = -fontHalfWidth, to = fontHalfWidth, center = 0}
            line[3] = {from = fontHalfWidth}
            self:updateWidth()
            return self.version, lineIdx, 0, (1 - (lineIdx % 2) * 2) * mfloor(lineIdx / 2) * lineHeight
        end
    
        local len = #line
        local toSpace, idx
        local min = line[1].to
        local max = line[len].from
        for i = 2, len - 1 do
            local space = line[i]
            if not space.fontString and mgte(space.to - space.from, fontWidth) then
                if not toSpace or mabs(space.center) < mabs(toSpace.center) then
                    toSpace = space
                    idx = i
                end
            end
        end
    
        if toSpace then
            if meq(toSpace.to - toSpace.from, toSpace.center) then
                toSpace.fontString = fontString
            else
                local newFrom = toSpace.center - fontHalfWidth
                local newTo = toSpace.center + fontHalfWidth
                tinsert(line, idx, {from = toSpace.from, to = newFrom, center = (newFrom + toSpace.from) / 2})
                tinsert(line, idx + 2, {from = newTo, to = toSpace.to, center = (toSpace.to + newTo) / 2})
                toSpace.fontString = fontString
                toSpace.from = newFrom
                toSpace.to = newTo
            end
    
            return self.version, lineIdx, toSpace.center, (1 - (lineIdx % 2) * 2) * mfloor(lineIdx / 2) * lineHeight
        else
            local newSpace
            if mabs(min) < mabs(max) then
                if mabs(min) + fontWidth > self.lineMaxEdge then
                    return self:insertFontStringToLine(lineIdx + 1, fontString, fontWidth, fontHalfWidth)
                end
    
                newSpace = {fontString = fontString, from = min - fontWidth, to = min, center = min - fontHalfWidth}
                line[1].to = newSpace.from
                tinsert(line, 2, newSpace)
            else
                if mabs(max) + fontWidth > self.lineMaxEdge then
                    return self:insertFontStringToLine(lineIdx + 1, fontString, fontWidth, fontHalfWidth)
                end
    
                newSpace = {fontString = fontString, from = max, to = max + fontWidth, center = max + fontHalfWidth}
                line[len].from = newSpace.to
                tinsert(line, len, newSpace)
            end
            
            self:updateWidth()
            return self.version, lineIdx, newSpace.center, (1 - (lineIdx % 2) * 2) * mfloor(lineIdx / 2) * lineHeight
        end
    end
    
    function anim:removeFontStringFromLine(fontString, version, lineIdx)
        if self.version ~= version or not lineIdx or not fontString then return end
        local line = self.lines[lineIdx]
        local preSpace = line and line[1]
        if not line or not preSpace then return end
        for i = 2, #line do
            local space = line[i]
            if space.fontString ~= fontString then
                preSpace = space
            else
                space.fontString = nil
                local nextSpace = line[i+1]
                if preSpace.fontString and nextSpace.fontString then
                elseif not preSpace.fontString and nextSpace.fontString then
                    preSpace.to = space.to
                    if preSpace.from then
                        preSpace.center = (preSpace.to + preSpace.from) / 2
                    end
                    tremove(line, i)
                elseif preSpace.fontString and not nextSpace.fontString then
                    nextSpace.from = space.from
                    if nextSpace.to then
                        nextSpace.center = (nextSpace.to + nextSpace.from) / 2
                    end
                    tremove(line, i)
                elseif not preSpace.fontString and not nextSpace.fontString then
                    preSpace.to = nextSpace.to
                    if preSpace.from and preSpace.to then
                        preSpace.center = (preSpace.to + preSpace.from) / 2
                    end
                    tremove(line, i)
                    tremove(line, i)
                    if not preSpace.from and not preSpace.to then
                        tremove(line, i - 1)
                    end
                end
                break
            end
        end
    
        local changed = false
        for i = #(self.lines), 3, -2 do
            if not self.lines[i][1] and not self.lines[i - 1][1] then
                self.lines[i] = nil
                self.lines[i - 1] = nil
                changed = true
            else
                break
            end
        end
        if changed then self:updateLineMaxEdge() end
        self:updateWidth()
    end

    anim:updateLineMaxEdge()

    return anim
end

local function verticalPath(elapsed, duration, distance)
    return 0, LibEasing.Linear(elapsed, 0, distance, duration)
end

local function powSizing(elapsed, duration, start, middle, finish)
    if elapsed < duration then
        local sep = duration * 0.667
        if elapsed < sep then
            return LibEasing.OutQuad(elapsed, start, middle - start, sep)
        else
            return LibEasing.InQuad(elapsed - sep, middle, finish - middle, duration - sep)
        end
    end
    return finish
end

function ClassicNFCT:AnimationOnUpdate()
    SetCVar("floatingCombatTextCombatDamage", self.db.global.blzDisabled and  "0" or "1")
    if animating:count() == 0 then
        animating:trunk()
        -- nothing in the animation list, so just kill the onupdate
        self.frame:SetScript("OnUpdate", nil)
        return
    end

    for fontString, _ in animating:iter() do
        local record = fontString.ClassicNFCT
        local elapsed = GetTime() - record.animatingStartTime
        if elapsed > record.animatingDuration
            or UnitIsDead(record.unit)
            or (not record.anchorFrame)
            or (not record.anchorFrame:IsShown())
        then
            self:recycleFontString(fontString)
        else
            local isTarget = UnitIsUnit(record.unit, "target")
            -- alpha
            local startAlpha = self.db.global.formatting.alpha
            if (self.db.global.useOffTarget and not isTarget) then
                startAlpha = self.db.global.offTargetFormatting.alpha
            end

            local alpha = LibEasing.InExpo(elapsed, startAlpha, -startAlpha, record.animatingDuration)
            fontString:SetAlpha(alpha)

            if alpha < startAlpha * 0.25 then
                self:recycleFontString(fontString, true)
            end

            -- sizing
            local scale = record.scale
            local critScale = 1
            if record.crit then
                if elapsed < record.animatingCritDuration then
                    critScale = powSizing(
                        elapsed, record.animatingCritDuration,
                        0.01, ANIMATION_CRIT_MAX_SCALE, ANIMATION_CRIT_SCALE
                    )
                else
                    critScale = ANIMATION_CRIT_SCALE
                end
            end

            local offsetScale = isTarget and self.db.global.formatting.scale or self.db.global.offTargetFormatting.scale
            local finalScale =  scale * offsetScale * critScale

            -- position
            local xOffset, yOffset = 0, 0
            -- if not record.crit then
                xOffset, yOffset = verticalPath(elapsed, record.animatingDuration, ANIMATION_VERTICAL_DISTANCE)
            -- end
            
            fontString:ClearAllPoints()
            fontString:SetParent(record.anchorFrame)
            fontString:SetScale(finalScale)
            local effScale = fontString:GetEffectiveScale()
            -- scale = scale * record.anchorFrame:GetScale()
            fontString:SetPoint(
                "CENTER", record.anchorFrame, "CENTER",
                -- (record.x + xOffset) / scale,
                (record.x + xOffset) / scale / critScale,
                self.db.global.distance / finalScale + (record.y + yOffset) / scale / critScale
                -- self.db.global.distance / offsetScale + (record.y + yOffset) * offsetScale
            )
        end
    end
end

function ClassicNFCT:Animate(fontString, record, anchorFrame, duration)
    record.animatingDuration = duration
    record.animatingCritDuration = duration * 0.333
    record.animatingStartTime = GetTime()
    record.anchorFrame = anchorFrame
    animating:emplace(fontString, true)

    -- start onupdate if it's not already running
    if (self.frame:GetScript("OnUpdate") == nil) then
        self.frame:SetScript("OnUpdate", function() self:AnimationOnUpdate() end)
    end
    self:AnimationOnUpdate()
end


----------------
-- FONTSTRING --
----------------
local function getFontPath(fontName)
    local fontPath = SharedMedia:Fetch("font", fontName)

    if (fontPath == nil) then
        fontPath = "Fonts\\FRIZQT__.TTF"
    end

    return fontPath
end

local fontStringCache = {}
function ClassicNFCT:getFontString()
    local fontString, record

    if (next(fontStringCache)) then
        fontString = tremove(fontStringCache)
    else
        fontString = self.frame:CreateFontString()
        fontString:SetFont(getFontPath(self.db.global.font), fontSize, self.db.global.fontFlag)
        fontString:SetParent(self.frame)
        -- fontString:SetIgnoreParentScale(false)
        fontString:SetIgnoreParentAlpha(true)
        fontString:SetDrawLayer("BACKGROUND")
        fontString.ClassicNFCT = {}
    end

    record = fontString.ClassicNFCT
    record.fontSize = fontSize
    record.fontFlag = self.db.global.fontFlag
    fontString:SetText("")
    fontString:SetAlpha(1)
    fontString:SetScale(1)
    if self.db.global.textShadow then
        fontString:SetShadowOffset(1, -1)
    else
        fontString:SetShadowOffset(0, 0)
    end
    fontString:Show()

    return fontString
end

function ClassicNFCT:recycleFontString(fontString, clearAnimOnly)
    local record = fontString.ClassicNFCT
    if record.anim then
        record.anim:removeFontStringFromLine(fontString, record.animVer, record.animIdx)
        record.anim = nil
    end

    if clearAnimOnly then return end

    animating:remove(fontString)
    fontString.ClassicNFCT = {}
    
    fontString:SetAlpha(0)
    fontString:ClearAllPoints()
    fontString:SetParent(self.frame)
    fontString:Hide()

    tinsert(fontStringCache, fontString)
end


----------
-- CORE --
----------
function ClassicNFCT:OnInitialize()
    -- setup db
    self.db = LibStub("AceDB-3.0"):New("ClassicNFCTDB", defaults, true)

    -- setup chat commands
    self:RegisterChatCommand("classicnfct", "OpenMenu")

    -- setup menu
    self:RegisterMenu()

    -- if the addon is turned off in db, turn it off
    if not self.db.global.enabled then
        self:Disable()
    end
    SetCVar("floatingCombatTextCombatDamage", self.db.global.blzDisabled and  "0" or "1")
end

function ClassicNFCT:OnEnable()
    playerGUID = UnitGUID("player")

    self:RegisterEvent("NAME_PLATE_UNIT_ADDED")
    self:RegisterEvent("NAME_PLATE_UNIT_REMOVED")
    self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
    self:RegisterEvent("PLAYER_TARGET_CHANGED")

    self.db.global.enabled = true
end

function ClassicNFCT:OnDisable()
    self:UnregisterAllEvents()

    for fontString, _ in animating:iter() do
        self:recycleFontString(fontString)
    end

    self.db.global.enabled = false
end


------------
-- EVENTS --
------------
function ClassicNFCT:NAME_PLATE_UNIT_ADDED(event, unitID)
    local guid = UnitGUID(unitID)

    unitToGuid[unitID] = guid
    guidToUnit[guid] = unitID
end

function ClassicNFCT:NAME_PLATE_UNIT_REMOVED(event, unitID)
    local guid = unitToGuid[unitID]

    unitToGuid[unitID] = nil
    guidToUnit[guid] = nil

	-- clear any fontStrings attachedk to this unit
	for fontString, _ in animating:iter() do
		if fontString.ClassicNFCT.unit == unitID then
			self:recycleFontString(fontString)
		end
    end
end

function ClassicNFCT:PLAYER_TARGET_CHANGED(event, unitID)
    self:AnimationOnUpdate()
end

function ClassicNFCT:CombatFilter(_, clue, _, sourceGUID, _, sourceFlags, _, destGUID, _, _, _, ...)
	if playerGUID == sourceGUID then -- Player events
		local destUnit = guidToUnit[destGUID]
		if destUnit then
			if (sfind(clue, "_DAMAGE")) then
				local spellID, spellName, spellSchool, amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing, isOffHand
				if (sfind(clue, "SWING")) then
					spellName, amount, overkill, _, resisted, blocked, absorbed, critical, glancing, crushing, isOffHand = "melee", ...
				elseif (sfind(clue, "ENVIRONMENTAL")) then
					spellName, amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing = ...
				else
					spellID, spellName, spellSchool, amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing, isOffHand = ...
                end
                if spellID == 75 then spellID, spellName = nil, "melee" end
				self:DamageEvent(destGUID, spellID, amount, school, critical, spellName)
			elseif(sfind(clue, "_MISSED")) then
				local spellID, spellName, spellSchool, missType, isOffHand, amountMissed

				if (sfind(clue, "SWING")) then
                    spellName, missType, isOffHand, amountMissed = "melee", ...
				else
					spellID, spellName, spellSchool, missType, isOffHand, amountMissed = ...
                end
                if spellID == 75 then spellID, spellName = nil, "melee" end
				self:MissEvent(destGUID, spellID, spellSchool, missType, spellName)
			end
		end
	elseif (xand(sourceFlags, COMBATLOG_OBJECT_TYPE_GUARDIAN) > 0 or xand(sourceFlags, COMBATLOG_OBJECT_TYPE_PET) > 0)	and xand(sourceFlags, COMBATLOG_OBJECT_AFFILIATION_MINE) > 0 then -- Pet/Guardian events
		local destUnit = guidToUnit[destGUID]
		if destUnit then
			if (sfind(clue, "_DAMAGE")) then
				local spellID, spellName, spellSchool, amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing, isOffHand
				if (sfind(clue, "SWING")) then
					spellName, amount, overkill, _, resisted, blocked, absorbed, critical, glancing, crushing, isOffHand = "pet", ...
				elseif (sfind(clue, "ENVIRONMENTAL")) then
					spellName, amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing = ...
				else
					spellID, spellName, spellSchool, amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing, isOffHand = ...
				end
				self:DamageEvent(destGUID, spellID, amount, "pet", critical, spellName)
			-- elseif(sfind(clue, "_MISSED")) then -- Don't show pet MISS events for now.
				-- local spellID, spellName, spellSchool, missType, isOffHand, amountMissed

				-- if (sfind(clue, "SWING")) then
					-- if destGUID == playerGUID then
					  -- missType, isOffHand, amountMissed = ...
					-- else
					  -- missType, isOffHand, amountMissed = "pet", ...
					-- end
				-- else
					-- spellID, spellName, spellSchool, missType, isOffHand, amountMissed = ...
				-- end
				-- self:MissEvent(destGUID, spellID, missType)
			end
		end
    end
end

function ClassicNFCT:COMBAT_LOG_EVENT_UNFILTERED()
    return ClassicNFCT:CombatFilter(CombatLogGetCurrentEventInfo())
end

-------------
-- DISPLAY --
-------------
function ClassicNFCT:textWithColor(text, school, isMelee)
    -- color text
    local textColor
    if self.db.global.damageColor then
        if school and DAMAGE_TYPE_COLORS[school] then
            textColor = DAMAGE_TYPE_COLORS[school]
        elseif isMelee then
            textColor = DAMAGE_TYPE_COLORS["melee"]
        else
            textColor = DAMAGE_TYPE_COLORS["spell"]
        end
    else
        if school and DAMAGE_TYPE_COLORS[school] or not isMelee then
            textColor = DAMAGE_TYPE_COLORS["spell"]
        else
            textColor = DAMAGE_TYPE_COLORS["melee"]
        end
    end
    return "|Cff".. textColor .. text .. "|r"
end

function ClassicNFCT:DamageEvent(guid, spellID, amount, school, crit, spellName)
    local text

    local unit = guidToUnit[guid]
    local icon = self.db.global.formatting.icon
    local isMelee = spellName == "melee"

    if (icon ~= "only") then
        local fmtStyle = self.db.global.fmtStyle
        if fmtStyle == "disable" then
            text = tostring(amount)
        else
            local abs = mabs(amount)
            local sym = amount >= 0 and "" or "-"
            if fmtStyle == "truncate" then
                local idx = 1
                local truncLen = #truncateWords
                while amount >= 1000 do
                    idx = idx + 1
                    if idx < truncLen and amount >= 1000 then
                        amount = amount / 1000
                    else
                        break
                    end
                end
                local word = truncateWords[idx]
                if idx == 1 then
                    text = tostring(amount)
                else
                    text = sformat("%.2f", amount)
                    local textLen = slen(text)
                    if textLen >= 6 then
                        text = ssub(text, 1, -4)
                    elseif textLen == 5 then
                        text = ssub(text, 1, -2)
                    end
                    if sfind(text, "%.00$") then
                        text = ssub(text, 1, -4)
                    elseif sfind(text, "%.%d0$") then
                        text = ssub(text, 1, -2)
                    end
                end
                text = text .. word
            elseif fmtStyle == "commaSep" then
                local abs = mabs(amount)
                local remain = abs % 1000
                local concat = {amount >= 0 and "" or "-", abs >= 1000 and sformat("%03d", remain) or remain}
                while abs >= 1000 do
                    abs = (abs - remain) / 1000
                    remain = abs % 1000
                    tinsert(concat, 2, ",")
                    tinsert(concat, 2, abs >= 1000 and sformat("%03d", remain) or remain)
                end
                text = tconcat(concat)
            end
        end

        -- color text
        text = self:textWithColor(text, school, isMelee)

        -- add icons
        if (icon ~= "none" and spellID) then
			local iconText = "|T"..GetSpellTexture(spellID)..":0|t"

			if (icon == "both") then
				text = iconText..text..iconText
			elseif (icon == "left") then
				text = iconText..text
			elseif (icon == "right") then
				text = text..iconText
			end
        end
    else
        -- showing only icons
        if (not spellID) then
            return
        end

        text = "|T"..GetSpellTexture(spellID)..":0|t"
    end

    self:DisplayText(guid, text, fontSize, crit, school == "pet", isMelee)
end

function ClassicNFCT:MissEvent(guid, spellID, spellSchool, missType, spellName)
    local text
    local unit = guidToUnit[guid]
    local icon = self.db.global.formatting.icon
    local isTarget = unit and UnitIsUnit(unit, "target")

    if (icon == "only") then
        return
    end

    local isMelee = spellName == "melee"
    text = self:textWithColor(MISS_EVENT_STRINGS[missType], spellSchool, isMelee)

    -- add icons
    if (icon ~= "none" and spellID) then
        local iconText = "|T"..GetSpellTexture(spellID)..":0|t"

        if (icon == "both") then
            text = iconText..text..iconText
        elseif (icon == "left") then
            text = iconText..text
        elseif (icon == "right") then
            text = text..iconText
        end
    end

    self:DisplayText(guid, text, fontSize, false, false, isMelee)
end

function ClassicNFCT:DisplayText(guid, text, size, crit, pet, melee)
    local fontString
    local unit = guidToUnit[guid]
    local nameplate

    if unit then nameplate = C_NamePlate.GetNamePlateForUnit(unit) end
    if not nameplate then return end

    if animating:count() > self.db.global.animations.animationCount then
        for fontString, _ in animating:iter() do
            self:recycleFontString(fontString)
            break
        end
    end

    fontString = self:getFontString()
    local record = fontString.ClassicNFCT

    record.text = text
    fontString:SetText(text)

    record.crit = crit
    record.pet = pet
    record.melee = melee

    lineHeight = lineHeight or (fontString:GetStringHeight() * ANIMATION_CRIT_SCALE + ANIMATION_VERTICAL_DISTANCE)

    record.scale = pet and self.db.global.petFormatting.scale
        or (melee and self.db.global.autoAttackFormatting.scale)
        or 1

    record.anim = createAnimation(guid)
    local width = fontString:GetUnboundedStringWidth() * record.scale * (crit and ANIMATION_CRIT_SCALE or 1) + ANIMATION_HORIZONTAL_PADDING
    record.animVer, record.animIdx, record.x, record.y = record.anim:insertFontStringToLine(1, fontString, width, width / 2)

    record.unit = unit
    record.guid = guid

    self:Animate(fontString, record, nameplate, self.db.global.animations.animationspeed)
end


-------------
-- OPTIONS --
-------------
local menu = {
    name = "ClassicNFCT",
    handler = ClassicNFCT,
    type = 'group',
    args = {
        enable = {
            type = 'toggle',
            name = L.UI["Enable"],
            desc = L.UI["If the addon is enabled"],
            get = "IsEnabled",
            set = function(_, newValue) if (not newValue) then ClassicNFCT:Disable() else ClassicNFCT:Enable() end end,
            order = 1,
            width = "half",
        },
        
        total = {
            type = 'group',
            name = '',
            order = 2,
            inline = true,
            width = 'full',
            disabled = function() return not ClassicNFCT.db.global.enabled end,
            args = {
                disableBlizzardFCT = {
                    type = 'toggle',
                    name = L.UI["Disable Blizzard FCT"],
                    desc = "",
                    get = function() return ClassicNFCT.db.global.blzDisabled end,
                    set = function(_, newValue)
                        ClassicNFCT.db.global.blzDisabled = newValue
                        SetCVar("floatingCombatTextCombatDamage", self.db.global.blzDisabled and  "0" or "1")
                    end,
                    order = 10,
                    width = "full",
                },

                animations = {
                    type = 'group',
                    name = L.UI["Animations"],
                    order = 30,
                    inline = true,
                    args = {
                        speed = {
                            type = 'range',
                            name = L.UI["Animation Speed"],
                            desc = L.UI["Default speed: 1"],
                            min = 0.5,
                            max = 3,
                            step = .01,
                            get = function() return ClassicNFCT.db.global.animations.animationspeed end,
                            set = function(_, newValue) ClassicNFCT.db.global.animations.animationspeed = newValue end,
                            order = 1,
                            width = "full",
                        },
                        lines = {
                            type = 'range',
                            name = L.UI["Animation Max Lines"],
                            desc = "",
                            min = 3,
                            max = 11,
                            step = 2,
                            get = function() return ClassicNFCT.db.global.animations.animMaxLines end,
                            set = function(_, newValue) ClassicNFCT.db.global.animations.animMaxLines = newValue end,
                            order = 2,
                            width = "full",
                        },
                        animationCount = {
                            type = 'range',
                            name = L.UI["Animation Count"],
                            desc = "",
                            min = 1,
                            max = 100,
                            step = 1,
                            get = function() return ClassicNFCT.db.global.animations.animationCount end,
                            set = function(_, newValue) ClassicNFCT.db.global.animations.animationCount = newValue end,
                            order = 3,
                            width = "full",
                        },
                    },
                },

                appearance = {
                    type = 'group',
                    name = L.UI["Appearance"],
                    order = 50,
                    inline = true,
                    args = {
                        font = {
                            type = "select",
                            dialogControl = "LSM30_Font",
                            name = L.UI["Font"],
                            order = 1,
                            values = AceGUIWidgetLSMlists.font,
                            get = function() return ClassicNFCT.db.global.font end,
                            set = function(_, newValue) ClassicNFCT.db.global.font = newValue end,
                        },
                        fontFlag = {
                            type = 'select',
                            name = L.UI["Font Flags"],
                            desc = "",
                            get = function() return ClassicNFCT.db.global.fontFlag end,
                            set = function(_, newValue) ClassicNFCT.db.global.fontFlag = newValue end,
                            values = fontFlags,
                            order = 2,
                        },
                        fontShadow = {
                            type = 'toggle',
                            name = L.UI["Text Shadow"],
                            get = function() return ClassicNFCT.db.global.textShadow end,
                            set = function(_, newValue) ClassicNFCT.db.global.textShadow = newValue end,
                            order = 3,
                        },
                        damageColor = {
                            type = 'toggle',
                            name = L.UI["Use Damage Type Color"],
                            desc = "",
                            get = function() return ClassicNFCT.db.global.damageColor end,
                            set = function(_, newValue) ClassicNFCT.db.global.damageColor = newValue end,
                            order = 4,
                        },
                    },
                },

                distance = {
                    type = 'group',
                    name = L.UI["Vertical Distance"],
                    order = 70,
                    inline = true,
                    args = {
                        scale = {
                            type = 'range',
                            name = "",
                            desc = L.UI["Vertical Distance of Text Starting from NamePlate"],
                            min = -100,
                            max = 100,
                            step = .01,
                            get = function() return ClassicNFCT.db.global.distance end,
                            set = function(_, newValue) ClassicNFCT.db.global.distance = newValue end,
                            order = 1,
                            width = "full",
                        },
                    }
                },

                formatting = {
                    type = 'group',
                    name = L.UI["Text Formatting"],
                    order = 90,
                    inline = true,
                    args = {
                        fmtStyle = {
                            type = "select",
                            name = L.UI["Format Style"],
                            values = fmtStyles,
                            get = function() return ClassicNFCT.db.global.fmtStyle end,
                            set = function(_, newValue) ClassicNFCT.db.global.fmtStyle = newValue end,
                            order = 1,
                        },
                        target = {
                            type = 'group',
                            name = L.UI["Target Text Appearance"],
                            order = 2,
                            inline = true,
                            args = {
                                icon = {
                                    type = 'select',
                                    name = L.UI["Icon"],
                                    desc = "",
                                    get = function() return ClassicNFCT.db.global.formatting.icon end,
                                    set = function(_, newValue) ClassicNFCT.db.global.formatting.icon = newValue end,
                                    values = iconValues,
                                    order = 51,
                                },
                                scale = {
                                    type = 'range',
                                    name = L.UI["Scale"],
                                    desc = "",
                                    min = .01,
                                    max = 5,
                                    step = .01,
                                    get = function() return ClassicNFCT.db.global.formatting.scale end,
                                    set = function(_, newValue) ClassicNFCT.db.global.formatting.scale = newValue end,
                                    order = 52,
                                },
                                alpha = {
                                    type = 'range',
                                    name = L.UI["Alpha"],
                                    desc = "",
                                    min = .01,
                                    max = 1,
                                    step = .01,
                                    get = function() return ClassicNFCT.db.global.formatting.alpha end,
                                    set = function(_, newValue) ClassicNFCT.db.global.formatting.alpha = newValue end,
                                    order = 53,
                                },
                            },
                        },
                        petScale = {
                            type = 'range',
                            name = L.UI["Pet Text Scale (Based on Target)"],
                            desc = "",
                            min = .01,
                            max = 5,
                            step = .01,
                            get = function() return ClassicNFCT.db.global.petFormatting.scale end,
                            set = function(_, newValue) ClassicNFCT.db.global.petFormatting.scale = newValue end,
                            order = 4,
                        },

                        autoAttackScale = {
                            type = 'range',
                            name = L.UI["Auto-Attack Text Scale (Based on Target)"],
                            desc = "",
                            min = .01,
                            max = 5,
                            step = .01,
                            get = function() return ClassicNFCT.db.global.autoAttackFormatting.scale end,
                            set = function(_, newValue) ClassicNFCT.db.global.autoAttackFormatting.scale = newValue end,
                            order = 5,
                        },

                        useOffTarget = {
                            type = 'toggle',
                            name = L.UI["Use Seperate Off-Target Text Appearance"],
                            desc = "",
                            get = function() return ClassicNFCT.db.global.useOffTarget end,
                            set = function(_, newValue) ClassicNFCT.db.global.useOffTarget = newValue end,
                            order = 100,
                            width = "full",
                        },
                        offTarget = {
                            type = 'group',
                            name = L.UI["Off-Target Text Appearance"],
                            hidden = function() return not ClassicNFCT.db.global.useOffTarget end,
                            order = 101,
                            inline = true,
                            args = {
                                scale = {
                                    type = 'range',
                                    name = L.UI["Scale"],
                                    desc = "",
                                    min = .01,
                                    max = 5,
                                    step = .01,
                                    get = function() return ClassicNFCT.db.global.offTargetFormatting.scale end,
                                    set = function(_, newValue) ClassicNFCT.db.global.offTargetFormatting.scale = newValue end,
                                    order = 2,
                                },
                                alpha = {
                                    type = 'range',
                                    name = L.UI["Alpha"],
                                    desc = "",
                                    min = 0,
                                    max = 1,
                                    step = .01,
                                    get = function() return ClassicNFCT.db.global.offTargetFormatting.alpha end,
                                    set = function(_, newValue) ClassicNFCT.db.global.offTargetFormatting.alpha = newValue end,
                                    order = 3,
                                },
                            },
                        },
                    },
                },
            },
        },
    },
}

function ClassicNFCT:OpenMenu()
    -- just open to the frame, double call because blizz bug
    InterfaceOptionsFrame_OpenToCategory(self.menu)
    InterfaceOptionsFrame_OpenToCategory(self.menu)
end

function ClassicNFCT:RegisterMenu()
    LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("ClassicNFCT", menu)
    self.menu = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("ClassicNFCT", "ClassicNFCT")
end
